#!@php_bin@
<?php

/*
 *       Copyright (c) 2010, Maarch SAS
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/**
 *
 * @version 0.1.0 beta
 * @author Bruno Carlin <bruno.carlin@maarch.org>
 */


#FIXME: interaction avec svn en cas de conflit...
#TODO: commands: patch
#TODO: commands: manage l10n
#TODO: accept a revision number


require("MaarchX/Cli/ArgsParser.php");


/********************
 ***    PARAMS    ***
 ********************/


$svnUrl = "http://svn.maarch.org";
$modules = Array("advanced_physical_archive", "alert_diffusion",
                 "attachments", "autofoldering", "basket", "cases",
                 "doc_converter", "doc_project", "entities", "folder",
                 "full_text", "indexing_searching", "ldap",
                 "notes", "notifications", "ocr_tesseract",
                 "physical_archive", "postindexing", "reports",
                 "templates", "web_service", "workflow");
$experimentalModules = Array("moreq", "life_cycle");
$current = "branches/1.1";
$dev = "trunk";
$suppArchives = Array(".tar.gz");


/*********************
 ***    HELPERS    ***
 *********************/


class ProcessControl
{
    public function __construct()
    {
        global $options;
        $this->descriptors = array(
            // stdin is a pipe that the child will read from
            0 => array("pipe", "php://stdin", "r"),
            // stdout is a pipe that the child will write to
            1 => array("pipe", "w"),
            // stderr is a file to write to
            2 => array("file", "tmp-err.log", "a"),
        );
        $this->stderr = "";
    }

    public function run($cmd)
    {
        $process = proc_open($cmd, $this->descriptors, $this->pipes);
        #while($el=stream_get_line($this->pipes[1], 8192)){
        while ($el = fgets($this->pipes[1], 8192)) {
            Main_msg($el);
        }
        proc_close($process);
        $tmp = file_get_contents('tmp-err.log');
        unlink('tmp-err.log');
        if ($tmp) {
            $this->stderr .= $tmp;
        }
    }

    public function getStderr()
    {
        return $this->stderr;
    }
}

function Main_printErr($msg)
{
    if ($msg) {
        $stderr = fopen("php://stderr", "w");
        fwrite($stderr, "\n**********************************************\n"
                        ."FOLLOWING ERRORS OCCURED!\n"
                        ."**********************************************\n"
                        ."\n"
                        . $msg
                        ."**********************************************\n");
        exit(1);
    }
}

function Main_msg($msg)
{
    global $options;
    if (!$options["quiet"]) {
        print $msg;
    }
}

function Main_removeDir($path)
{
    if (is_file($path)) {
        unlink($path);
    } else {
        foreach (glob($path."/*") as $subpath) {
            Main_removeDir($subpath);
        }
        foreach (glob($path."/.*") as $subpath) {
            $lastPart = explode("/", $subpath);
            $lastPart = $lastPart[count($lastPart) - 1];
            if (!in_array($lastPart, array('.', '..'))) {
                Main_removeDir($subpath);
            }
        }
        rmdir($path);
    }
}


/******************
 ***    MAIN    ***
 ******************/


function Main_main()
{
    global $options, $dev;
    #php build.php [-q] checkout|co [--branch=trunk] [PATH]
    #php build.php [-q] update|up [PATH]

    ini_set('track_errors', '1');
    
    $brchDesc = "\"dev\" and \"trunk\" are aliases for the "
              . "development version, \"current\", \"stable\" and "
              . "\"latest\" are aliases for the latest stable version";
    $checkoutCmdParser = new MaarchX_Cli_ArgsParser();
    $checkoutCmdParser->addArg("branch",
                               array(
                                   "short" => "b",
                                   "long" => "branch",
                                   "default" => $dev,
                                   "help" => "Branch to checkout." . $brchDesc,
                               ));
    $checkoutCmdParser->addArg("apps",
                               array(
                                   "short" => "a",
                                   "long" => "apps",
                                   "default" => "maarch_entreprise",
                                   "help" => "Application to checkout",
                               ));
    $checkoutCmdParser->addSwitch(
        "experimental",
        array(
            "short" => "e",
            "long" => "with-experimental",
            "help" => "Checkout experimental modules too",
        )
    );
    $checkoutCmdParser->addPositional(
        "path", 
        array(
            "default" => getcwd(),
            "help" => "Path the working copy will be checked out to",
        )
    );

    $updateCmdParser = new MaarchX_Cli_ArgsParser();
    $updateCmdParser->addPositional(
        "path", 
        array(
            "default" => getcwd(),
            "help" => "Path of the working copy to update",
        )
    );

    $sqlCmdParser = new MaarchX_Cli_ArgsParser();
    $sqlCmdParser->addPositional(
        "path", 
        array(
            "default" => getcwd(),
            "help" => "Path of the working copy for wich SQL "
                      ."scripts must be generated",
        )
    );
    $sqlCmdParser->addArg(
        "dbengine",
        array(
            "short" => "d",
            "long" => "dbengine",
            "default" => "postgresql",
            "help" => "Database engine",
        )
    );

    $keygenCmdParser = new MaarchX_Cli_ArgsParser();
    $keygenCmdParser->addPositional(
        "path",
        array(
            "default" => getcwd(),
            "help" => "Path of the working copy for wich public and"
                      ." private key files must be generated",
        )
    );
    $keygenCmdParser->addArg("destination",
                             array(
                                 "short" => "d",
                                 "long" => "dest",
                                 "mandatory" => true,
                                 "help" => "Destination of public and"
                                           . " private key files",
                             ));


    $distCmdParser = new MaarchX_Cli_ArgsParser();
    $distCmdParser->addArg("tmppath",
                           array(
                               "short" => "t",
                               "long" => "tmp-dir",
                               "default" => getcwd(),
                               "help" => "Create the tmp dir here",
                           ));
    $distCmdParser->addPositional(
        "filename",
        array(
            "mandatory" => true,
            "help" => "Path and name of the package",
        )
    );
    $distCmdParser->addPositional(
        "path", 
        array(
            "default" => getcwd(),
            "help" => "Path of the working copy to package",
        )
    );

    $branchCmdParser = new MaarchX_Cli_ArgsParser();
    $branchCmdParser->addArg("branch",
                             array(
                                 "short" => "b",
                                 "long" => "branch",
                                 "default" => $dev,
                                 "help" => "The path to branch." . $brchDesc,
                             ));
    $branchCmdParser->addPositional("branch_name",
                                    array(
                                        "mandatory" => true,
                                        "help" => "Name of the new branch",
                                    ));

    $tagCmdParser = new MaarchX_Cli_ArgsParser();
    $tagCmdParser->addArg("branch",
                          array(
                              "short" => "b",
                              "long" => "branch",
                              "default" => $dev,
                              "help" => "The path to tag." . $brchDesc,
                          ));
    $tagCmdParser->addPositional("tag_name", 
                                 array(
                                     "mandatory" => true,
                                     "help" => "Name of the new tag",
                                 ));

    $argParser = new MaarchX_Cli_ArgsParser();
    $argParser->addSwitch("quiet",
                          array(
                              "short" => "q",
                              "long" => "quiet",
                              "help" => "Disable all output by the program",
                          ));
    $argParser->addCommand("checkout", Array("co"), $checkoutCmdParser,
                           "Create a new working copy of the application");
    $argParser->addCommand("update", Array("up"), $updateCmdParser,
                           "Perform a recursive update on the working copy");
    $argParser->addCommand("dist", Array(), $distCmdParser,
                           "Create a package of the application");
    $argParser->addCommand("tag", Array(), $tagCmdParser,
                           "Create a new tag for the application");
    $argParser->addCommand("branch", Array(), $branchCmdParser,
                           "Create a new branch for the application");
    $argParser->addCommand("sql", Array(), $sqlCmdParser,
                           "Generate SQL scripts for the application");
    $argParser->addCommand("keygen", Array(), $keygenCmdParser,
                           "Generates public and private key file");
    $options = $argParser->parseArgs($GLOBALS["argv"]);
    if ($options === false) {
        exit;
    }

    if (!array_key_exists("command", $options)) {
        Main_msg("Aucune commande n'a été fournie !");
        print $argParser->usage();
        exit();
    }

    Main_dispatch($options["command"]);
}

function Main_dispatch($command)
{
    if (function_exists("Main_do".ucfirst($command["name"]))) {
        call_user_func("Main_do".ucfirst($command["name"]), $command["opts"]);
    }
}


/*********************
 ***    ACTIONS    ***
 *********************/

function Main_doCheckout($opts)
{
    global $svnUrl, $modules, $current, $dev;

    switch ($opts['branch']) {
        case "dev":
        case "trunk":
            $branch = $dev;
            break;

        case "current":
        case "latest":
        case "stable":
            $branch = $current;
            break;

        default:
            $branch = $opts['branch'];
    }

    if ($opts["experimental"] && $branch == $dev) {
        global $experimentalModules;
        array_splice($modules, count($modules), 0, $experimentalModules);
    }

    $pc = new ProcessControl();
    Main_msg("\n* checking out core\n");
    $pc->run(sprintf("svn co \"%s/core/%s\" \"%s\"", 
                     $svnUrl, $branch, $opts["path"]));
    Main_msg("\n* checking out apps\n");
    $pc->run(sprintf("svn co \"%s/%s/%s\" \"%s/apps/%s\"", 
                     $svnUrl, $opts["apps"], $branch, 
                     $opts["path"], $opts["apps"]));

    foreach ($modules as $module) {
        Main_msg("\n* checking out module $module\n");
        $pc->run(sprintf("svn co \"%s/%s/%s\" \"%s/modules/%s\"", 
                         $svnUrl, $module, $branch, $opts["path"], $module));
    }
    Main_printErr($pc->stderr);
}

function Main_doUpdate($opts)
{

    chdir($opts["path"]);

    $pc = new ProcessControl();

    Main_msg("* updating core\n");
    $pc->run("svn up .");

    Main_msg("\n* updating \"apps\"\n");
    $pc->run("svn up apps/*");

    foreach (glob("modules/*") as $module) {
        Main_msg("\n* updating \"$module\"\n");
        $pc->run("svn up \"$module\"");
    }
    Main_printErr($pc->stderr);
}

function Main_doDist($opts)
{
    global $suppArchives;

    $filepath = $opts['filename'];
    $filename = explode(DIRECTORY_SEPARATOR, $filepath);
    $filename = $filename[count($filename) - 1];
    foreach ($suppArchives as $filetype) {
        if (strrpos($filetype, $filename)) {
            $filetype = $filetype;
            break;
        }
    }
    $basename = str_replace($filetype, "", $filename);
    $tmpdir = $opts["tmppath"]."/".$basename;
    $wcpath = realpath($opts["path"]);

    mkdir($tmpdir, 0777, True);

    chdir($wcpath);

    $pc = new ProcessControl();

    $pc->run(sprintf("svn export \"%s\" \"%s/\"", '.', $tmpdir));

    mkdir($tmpdir.'/apps');
    $apps = scandir("apps");
    $apps = $apps[2];
    $pc->run(sprintf("svn export \"apps/%s\" \"%s/%s/apps/%s\"",
                     $apps, $tmpdir, $basename, $apps));

    mkdir($tmpdir.'/modules');
    foreach (glob("modules/*") as $module) {
        $pc->run(sprintf("svn export \"%s\" \"%s/%s/%s\"",
                         $module, $tmpdir, $basename, $module));
    }

    chdir($tmpdir);
    switch($filetype) {
        case ".tar.gz":
            $pc->run("tar czf $filepath $basename");
            break;

        case ".tar.bz2":
            $pc->run("tar cjf $filepath $basename");
            break;

        case ".tar.lzma":
            $pc->run("tar -cf --lzma $filepath $basename");
            break;

        case ".tar.xz":
            $pc->run("tar -cf --xz $filepath $basename");
            break;

        case ".zip":
            $pc->run("zip -r $filepath $basename");
            break;

        default:
            rename($basename, $filepath);
            break;
    }


    chdir($wcpath);

    Main_removeDir($tmpdir);

    Main_printErr($pc->stderr);
}


function Main_doBranch($opts)
{
    global $svnUrl, $dev, $current, $modules, $experimentalModules;

    switch ($opts['branch']) {
        case "dev":
        case "trunk":
            $branch = $dev;
            break;

        case "current":
        case "latest":
        case "stable":
            $branch = $current;
            break;

        default:
            $branch = $opts['branch'];
    }

    array_splice($modules, count($modules), 0, $experimentalModules);

    $pc = new ProcessControl();
    $pc->run(sprintf("svn branch \"%s/core/%s\" \"%s/core/branches/%s\"",
                     $svnUrl, $branch, $svnUrl, $opts["branch_name"]));
    $pc->run(sprintf("svn branch \"%s/maarch_entreprise/%s\" "
                     ."\"%s/maarch_entreprise/branches/%s\"",
                     $svnUrl, $branch, $svnUrl, $opts["branch_name"]));
    foreach ($modules as $module) {
        $pc->run(sprintf("svn branch \"%s/%s/%s\" \"%s/%s/branches/%s\"",
                         $svnUrl, $module, $branch, $svnUrl, $module,
                         $opts["branch_name"]));
    }
}


function Main_doTag($opts)
{
    global $svnUrl, $dev, $current, $modules, $experimentalModules;

    switch ($opts['branch']){
        case "dev":
        case "trunk":
            $branch = $dev;
            break;

        case "current":
        case "latest":
        case "stable":
            $branch = $current;
            break;

        default:
            $branch = $opts['branch'];
    }

    array_splice($modules, count($modules), 0, $experimentalModules);

    $pc = new ProcessControl();
    $pc->run(sprintf("svn tag \"%s/core/%s\" \"%s/core/tags/%s\"",
                     $svnUrl, $branch, $svnUrl, $opts["tag_name"]));
    $pc->run(sprintf("svn tag \"%s/maarch_entreprise/%s\" "
                     ."\"%s/maarch_entreprise/tags/%s\"",
                     $svnUrl, $branch, $svnUrl, $opts["tag_name"]));
    foreach ($modules as $module) {
        $pc->run(sprintf("svn tag \"%s/%s/%s\" \"%s/%s/tags/%s\"",
                         $svnUrl, $module, $branch, $svnUrl, $module,
                         $opts["tag_name"]));
    }
}

function Main_doSql($opts)
{

    chdir($opts["path"]);

    $pc = new ProcessControl();

    $datafile = fopen("data.sql", "w");
    $structurefile = fopen("structure.sql", "w");
    
    $filename = "core/sql/data/core.${opts['dbengine']}.sql";
    if (file_exists($filename)) {
        fwrite($datafile, "\n\n-- $filename\n\n");
        fwrite($datafile, file_get_contents($filename));
    }
    
    $filename = "core/sql/structure/core.${opts['dbengine']}.sql";
    if (file_exists($filename)) {
        fwrite($structurefile, "\n\n-- $filename\n\n");
        fwrite($structurefile, file_get_contents($filename));
    }


    foreach (glob("modules/*") as $module) {
        $moduleName = explode("/", $module);
        $moduleName = $moduleName[1];
        
        $filename = "$module/sql/data/$moduleName.${opts['dbengine']}.sql";
        if (file_exists($filename)) {
            fwrite($datafile, "\n\n-- $filename\n\n");
            fwrite($datafile, file_get_contents($filename));
        }
        
        $filename = "$module/sql/structure/$moduleName.${opts['dbengine']}.sql";
        if (file_exists($filename)) {
            fwrite($structurefile, "\n\n-- $filename\n\n");
            fwrite($structurefile, file_get_contents($filename));
        }
    }


    $apps = scandir("apps");
    $apps = $apps[2];
    $filename = "apps/$apps/sql/data/apps.${opts['dbengine']}.sql";
    if (file_exists($filename)) {
        fwrite($datafile, "\n\n-- $filename\n\n");
        fwrite($datafile, file_get_contents($filename));
    }
    
    $filename = "apps/$apps/sql/structure/apps.${opts['dbengine']}.sql";
    if (file_exists($filename)) {
        fwrite($structurefile, "\n\n-- $filename\n\n");
        fwrite($structurefile, file_get_contents($filename));
    }

    $errorMsg = $pc->stderr;
    if (strlen($errorMsg)) {
        Main_printErr($errorMsg);
    } else {
        Main_msg("the files data.sql and structure.sql have been "
                 . "generated successfully!\n");
    }

}

/**
* Generates private and public keys
*/
function Main_doKeygen($opts) 
{
    $pathToPublicKey = $opts['destination'].DIRECTORY_SEPARATOR."maarchkey.pub";
    $pathToPrivateKey = $opts['destination'].DIRECTORY_SEPARATOR."maarchkey";

    if (!is_writable($opts['destination'])) {
        print $opts['destination'] . " directory does not exist or is not "
              . "writable\n";
        exit(1);
    }

    $privateKey = openssl_pkey_new(
        array(
            'private_key_bits' => 1024,
            'private_key_type' => OPENSSL_KEYTYPE_RSA,
        )
    );
    openssl_pkey_export_to_file($privateKey, $pathToPrivateKey);
    $keyDetails = openssl_pkey_get_details($privateKey);
    file_put_contents($pathToPublicKey, $keyDetails['key']);
}

Main_main();
